/***********************************************************************

  webtty.js - client side script used to display output from server and 
              to send keypresses to server

  This file is part of the webtty package. The webtty package was
  written by Martin Steen Nielsen

  Development is hosted by testape.com at www.testape.com/webtty

  webtty is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  webtty is distributed in the hope that it will be useful,but 
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with webtty; if not, write to the Free Software Foundation, 
  Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

***********************************************************************/
var webTtyTaskArray= new Array;


function fullURIEncode(key)
{
    
  // return encodeURICOmponent(key);
    if  (window.dictionary.alphaRev[key])
	return window.dictionary.alphaRev[key];

    if (key>255)
	return "%"+("00" + (key>>8).toString(16)).match(/([0-9A-Fa-f][0-9A-Fa-f])$/)[0]
	+ "%"+("00" + (key & 255).toString(16)).match(/([0-9A-Fa-f][0-9A-Fa-f])$/)[0];
    else
	return "%"+("00" + key.toString(16)).match(/([0-9A-Fa-f][0-9A-Fa-f])$/)[0];
}


/***********************************************************************
   closes terminal
***********************************************************************/
function wclose(id)
{
  var elem = document.getElementById(id);
  elem.style.display='none';
  return false;
}




/***********************************************************************
   constructor webtty is called from client. Constructor expects to 
   receive id of element that will hold the html for the new process. 
   Function exit_func(err) is optional, but if given, it will be called 
   on errors or if the server terminates the process early.

   Sample usage -->

   <html>
     <script type="text/javascript" src="webtty.js" charset="UTF-8">
     <body onLoad="new webtty('webtty_element');">
       <div id="webtty_element"></div>
     </body>
   </html>


***********************************************************************/

function zeroOrText(s){
  if (s == "\0")
            return "";
  else
    return s;
};

function newStrLen(l) {
    var s='';
    while (l>0) {
        s = s +' ';
        l--;
    }
    return s;
};

var capPaste = false;

var TTY_STATE_INIT = 0;
var TTY_STATE_RUNNING = 1;
var TTY_STATE_CLOSING = 2;
var TTY_STATE_RESERVED = 3;
var TTY_STATE_CLOSED = 4;

function webtty(receiving_elementID, exit_func)
{
    this.getRange = function () {
        var ctrl = this.output_area;
        if(ctrl.setSelectionRange)
	      {
		        ctrl.focus();
		        ctrl.setSelectionRange(pos,pos);
	      }
	      else if (ctrl.createTextRange) {
		        var range = ctrl.createTextRange();
		        range.collapse(true);
		        range.moveEnd('character', pos);
		        range.moveStart('character', pos);
		        range.select();
	      }
    };
    this.setCursorTo = function (pos) {
        var ctrl = this.output_area;
        if(ctrl.setSelectionRange)
	      {
		        ctrl.focus();
		        ctrl.setSelectionRange(pos,pos);
	      }
	      else if (ctrl.createTextRange) {
		        var range = ctrl.createTextRange();
		        range.collapse(true);
		        range.moveEnd('character', pos);
		        range.moveStart('character', pos);
		        range.select();
	      }
    };
    this.setCursor = function (){
	var new_cell = this.tty.fn.eFindCell(this.tty, this.tty.offset + this.tty.y, this.tty.x);
	if(this.tty.cursor.cell != new_cell) {
	    $(this.tty.cursor.cell).toggleClass('ttycursor', false);
	    $(new_cell).toggleClass('ttycursor', true);
	    this.tty.cursor.cell = new_cell;
	}
/*
        if((this.tty.offset + this.tty.y != this.tty.cursor.y) || ( this.tty.x != this.tty.cursor.x) ) {
            if(this.tty.cursor.y>this.tty.offset) {
                var old_cell = this.tty.fn.eFindCell(this.tty, this.tty.cursor.y, this.tty.cursor.x);
                $(old_cell).toggleClass('ttycursor', false);
            }
            var new_cell = this.tty.fn.eFindCell(this.tty, this.tty.offset + this.tty.y, this.tty.x);
            $(new_cell).toggleClass('ttycursor', true);
            this.tty.cursor.y = this.tty.offset + this.tty.y;
            this.tty.cursor.x = this.tty.x;
        }*/

    }

    this.tty = {
	'state' : 0, // 0-init 1-running  2-closing  3-reserved 4-closed 
	'data' : '',
	'autoscroll' : true,
	'timeout' : 200,
	'emptyRow' : "<div class='ttyrow'><div class='ttycell'>&nbsp;</div></div>",
        'node' : this,
        'font' : {
            'underlined' : false,
            'blink' : false,
            'inverse' : false,
            'bold' : false,
            'invisible' : false,
            'color' : "white",
            'bgcolor' : "black"
        },
        'cache' : {
            'rows' : new Array()
        },
        'tab' : {},
        'x' : 1,
        'y' : 1,
	'relative' : false,
        'cursor' : {
            'x' : 1,
            'y' : 1,
	    'cell' : null
        },
        'offset' : 0,
        'max_x' : 80,
        'max_y' : 24,
        'scroll':{
            'top': 1,
            'bottom' : 24
        },
        'wrap' : true,
        'fn':{
            'shift': function (tty, top, size, delta) {
		
            },
            'del'  : function (tty, top, size){
		
            },
            'clear' :function (tty, top, size) {
		tty.fn.eFindRow(tty, top + size-1);
		$("div.ttyrow",tty.node.ttyel).each(
                    function(i,el){
                        if (i>=top-1 && i < top-1+size) {
                            $(el).html('<div class="ttycell">&nbsp;</div>');
			    tty.fn.rebuildCols(tty, i+1)
			}
                    }
                );
            },
            'nil'  : function (tty, re_a) {
            },
            'clearLineFromPos': function (tty) {
                $("div.ttycell",tty.fn.eFindRow(tty, tty.y + tty.offset)).each(
                    function(i,el){
                        if (i >= tty.x-1) {
                            if(i==0)
                                $(el).html("&nbsp;");
                            else
                                $(el).remove();
                        }
                    }
                );
                tty.fn.rebuildCols(tty, tty.y + tty.offset);
            } ,
            'clearLineBeforePos': function (tty) {
                $("div.ttycell",tty.fn.eFindRow(tty, tty.y + tty.offset)).each(
                    function(i,el){
                        if (i < tty.x) {
                            $(el).html("&nbsp;");
                        }
                    }
                );
                tty.fn.rebuildCols(tty, tty.y + tty.offset);
            },
            'clearLine':function (tty) {
                tty.fn.clear(tty, tty.y + tty.offset, 1);
            },
            'clearScreenFromPos': function (tty) {
                tty.fn.clearLineFromPos(tty);
                tty.fn.clear(tty, tty.offset + tty.y + 1, tty.max_y - tty.y );
            },
            'clearScreenBeforePos': function (tty) {
                tty.fn.clearLineBeforePos(tty);
                tty.fn.clear(tty, tty.offset + 1, tty.y-1);
            },
            'clearScreen': function (tty) {
                tty.fn.clear(tty, tty.offset+1, tty.max_y)
            },

            'saveCursorPos':function (tty) {
                tty.save_pos = {'x' : tty.x, 'y' : tty.y};
            },
            'restoreCursorPos':function (tty) {
                tty.x = tty.save_pos.x;
                tty.y = tty.save_pos.y;
            },
	    'fillScreenWithE' : function (tty, re_a) {
		for(var y = 1; y<= tty.max_y; y++)
		    for(var x = 1; x<= tty.max_x; x++) {
			tty.x = x;
			tty.y = y;
			tty.fn.eSet(tty, 'E');
		    }
            },
            'setCursorPos' : function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                tty.y = n + (tty.relative ? tty.scroll.top - 1 : 0);
                tty.x = Number(re_a[3] ? re_a[3] : 1);
            },
	    'Return' : function (tty) {
		tty.x = 1;
	    },
	    'lineFeed' : function (tty) { 
		if(tty.y < tty.scroll.bottom ) {
		    tty.y = tty.y+1; 
		} else {
		    tty.fn.scrollRegionUp(tty, 1);
		}
		
	    },

            'moveCursorUp'  : function (tty, re_a) {
		var N = Number(typeof(re_a[1]) != 'undefined' ? re_a[1] : 1);
                tty.y = tty.y - N;
		var top = tty.relative ? tty.scroll.top : 1;
                tty.y = tty.y < top ? top : tty.y;
            }, 

            'moveCursorDown' :function (tty, re_a) {
                var N = Number(typeof(re_a[1]) != 'undefined' ? re_a[1] : 1);
		tty.y = tty.y + N;
		var bottom = tty.relative ? tty.scroll.bottom : tty.max_y;
               	tty.y = tty.y > bottom ? bottom : tty.y;
            },

            'moveCursorToRow'  : function (tty, re_a) {
                var N = Number(re_a[1] ? re_a[1] : 1);
		N = N + (tty.relative ? tty.scroll.top - 1 : 0 );
                if (tty.y >= N) {
                    re_a[1] =  tty.y - N;
                    tty.fn.moveCursorUp(tty, re_a);
                } else {
                    re_a[1] =  N - tty.y;
                    tty.fn.moveCursorDown(tty, re_a);
                };
            }, 
	   
	    'wrap' : function (tty) {
		if (tty.x>tty.max_x) {
		    if(tty.wrap) {
			tty.fn.Return(tty);
			tty.fn.lineFeed(tty);
		    } else {
		    	tty.x = tty.max_x;
		    }
		}
		if ( tty.x < 1) {
		    if (tty.wrap) {
			if(tty.y>tty.scroll.top) {
			    tty.y = tty.y - 1;
			    tty.x = tty.max_x;
			} else {
			    tty.x = 1;
			}
			
		    }
		    else {
			tty.x = 1;
		    }
		}
	    },

            'moveCursorRight' :function (tty, re_a) {
                var n = Number(typeof(re_a[1]) != 'undefined' ? (re_a[1]>0 ? re_a[1] : 1) : 1);
		while(n>0) {
		    n--;
		    tty.x = tty.x + 1;
		    tty.fn.wrap(tty);
		}
            },
            'moveCursorToColumn' :function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                tty.x = n;
                tty.x = Number(tty.x > tty.max_x ? tty.max_x :tty.x); //!!!check tty!!!
            },
            'moveCursorLeft' :function (tty, re_a) {
                var n = Number(typeof(re_a[1]) != 'undefined' ? (re_a[1]>0 ? re_a[1] : 1) : 1);
		while(n>0) {
		    n --;
                    tty.x = tty.x - 1;
		    tty.x = tty.x <1 ? 1: tty.x;
//		    tty.fn.wrap(tty);
		}
            },
            'insertLine' :function (tty, re_a) {
                var n = Number(re_a[1] ? re_a[1] : 1);
                if(tty.scroll.bottom - tty.y - n > 0) {
                    tty.fn.shift(tty,tty.offset + tty.y - 1,tty.scroll.bottom - tty.y - n + 1, n);
                } else {
                    tty.fn.clear(tty,tty.offset + tty.y - 1,tty.scroll.bottom - tty.y);
                }
            },
            'deleteLine' :function (tty, re_a) {
                var N = Number(re_a[1] ? re_a[1] : 1);
                tty.fn.clear(tty,tty.offset + tty.y, N);
            },
            'deleteChar' :function (tty, re_a) {
                var N = Number(re_a[1] ? re_a[1] : 1);
		$("div.ttycell",tty.fn.eFindRow(tty, tty.y + tty.offset)).each(
                    function(i,el){
                        if (i>=tty.x - N && i < tty.x) {
			    $(el).remove();
                        }
                    }
                );
                tty.fn.rebuildCols(tty, tty.y + tty.offset);		
            },
            'eraseChar' :function (tty, re_a) {
                var N = Number(re_a[1] ? re_a[1] : 1);
                while (N>0) {
                    //TODO
                    N--;
                }
            },
            'setScrollRegion' :function (tty, re_a) {
                var top = Number(re_a[2] ? re_a[2] : 1);
                var bottom = Number(re_a[4] ? re_a[4] : tty.max_y);
                tty.scroll.top = top;
                tty.scroll.bottom = bottom;
            },
            'scrollRegionDown' :function (tty, N) {
		while(N>0) {
		    N--;
		    $(tty.cache.rows[tty.offset+tty.scroll.top - 1]).before(tty.emptyRow);     
		    $(tty.cache.rows[tty.offset+tty.scroll.bottom -1]).remove();
		   
		    tty.fn.rebuildRows(tty, tty.offset+tty.scroll.top -1);  
		}
            },
            'scrollRegionUp' :function (tty, N) {
		while(N>0) {
		    N--;
		    if ( tty.scroll.top >1) {
			$(tty.cache.rows[tty.offset + tty.scroll.top - 1]).remove();
		    }
		    //  if ( tty.scroll.bottom < tty.max_y) {
		    $(tty.cache.rows[tty.offset+tty.scroll.bottom - 1]).after(tty.emptyRow);
		    //   }
		    var el = $(tty.cache.rows[tty.offset+tty.scroll.bottom - 1]).next()[0];

		    if ( tty.scroll.top == 1) {
			tty.offset ++;
		    }

		    tty.cache.rows.push(el);
		    tty.fn.rebuildCols(tty, tty.cache.rows.length);
		    //tty.fn.rebuildRows(tty);  
		}
            },
	    'getTabShift' : function (tty) {
		var pos8 = (8 - ((tty.x) % 8)) + tty.x;
		pos8 = pos8 < tty.max_x ? pos8 : tty.max_x;
		var pos = 7777777;
		for (var i in tty.tab) {
		    pos = (Number(i) > tty.x & Number(i) < pos) ? Number(i) : pos;
		}
		return (pos != 7777777 ? pos : pos8 ) - tty.x - 1;
	    },
	    'clearTab' : function (tty, re_a) {
		var n = Number(re_a[1] ? re_a[1] : 0);
		switch (n) {
		case 0: 
		    delete tty.tab[tty.x]; 
		    break;
		case 3:
		    tty.tab = {};
		    break;
		default:
		}
	    },
            'resetTerminal' : function (tty) {/*  TODO reset terminal*/;
                tty.x = 1;
                tty.y = 1;
		tty.fn.saveCursorPos(tty);
		tty.lnm = false;
//                tty.max_x = 80;
                tty.max_y = 24;
		tty.relative = false;
                tty.scroll.top = 1;
                tty.scroll.bottom = 24;
                tty.wrap = true;
		tty.offset = 0;	
		tty.inverse = false;
		tty.fn.defaultFont(tty);
		$("div.ttyrow",tty.node.ttyel).each( 
		    function(i,el){
			$(el).remove();
		    });
		tty.fn.rebuildRows(tty);
		tty.fn.clearTab(tty, {1:3});
		
		tty.fn.eFindCell(tty, tty.max_y, tty.max_x);
		if (tty.autoscroll)
		    tty.node.ttyel.scrollTop = tty.node.ttyel.scrollHeight - tty.node.ttyel.clientHeight;
            },
	    'defaultFont' : function(tty) {
                tty.font.bold = false;
                tty.font.underlined = false;
                tty.font.blink = false;
                tty.font.inverse = false;
                tty.font.invisible = false;
                tty.font.color = "#252525";//"white";
                tty.font.bgcolor = "#cacaca";//"black";
            },
            'setColorParams' :function (tty, re_a) {
                var r = /[0-9]+/g;
                var N = r.exec(re_a[1]);
		
                if (! N)
                    tty.fn.defaultFont(tty);
                while( N ) {
                    switch(parseInt(N[0],10)) {
                    case 0://normal
                        tty.fn.defaultFont(tty);
                        break;
                    case 1:
                        tty.font.bold = true;
                        break;
                    case 2:
                        tty.font.bold = true;
                        break
                    case 4:
                        tty.font.underlined = true;
                        break;
                    case 5:
                        tty.font.blink = true;
                        break;
                    case 7:
                        tty.font.inverse = true;
                        break;
                    case 8:
                        tty.font.invisible = true;
                        break;
                    case 22:
                        tty.font.bold = false;
                        break;
                    case 24:
                        tty.font.underlined = false;
                        break;
                    case 25:
                        tty.font.blink = false;
                        break;
                    case 27:
                        tty.font.inverse = false;
                        break;
                    case 28:
                        tty.font.invisible = false;
                        break;
                    case 30:
                        tty.font.color = "#252525";//"black";
                        break;
                    case 31:
                        tty.font.color = "#ba7373";//"red";
                        break;
                    case 32:
                        tty.font.color = "#739456";//"green";
                        break;
                    case 33:
                        tty.font.color = "#a4712e";//"yellow";
                        break;
                    case 34:
                        tty.font.color = "#25414f";//"blue";
                        break;
                    case 35:
                        tty.font.color = "#6f4f70";//"magenta";
                        break;
                    case 36:
                        tty.font.color = "#3c6f69";//"cyan";
                        break;
                    case 37:
                        tty.font.color = "#ffffff";//"white";
                        break;
                    case 39://default
                        tty.font.color = "#252525";//"white";
                        break;
                    case 40:
                        tty.font.bgcolor = "#666666";//"black";
                        break;
                    case 41:
                        tty.font.bgcolor = "#c14b4b";//"red";
                        break;
                    case 42:
                        tty.font.bgcolor = "#6ca318";//"green";
                        break;
                    case 43:
                        tty.font.bgcolor = "#7f672e";//"yellow";
                        break;
                    case 44:
                        tty.font.bgcolor = "#537595";//"blue";
                        break;
                    case 45:
                        tty.font.bgcolor = "#a971a7";//"magenta";
                        break;
                    case 46:
                        tty.font.bgcolor = "#70999a";//"cyan";
                        break;
                    case 47:
                        tty.font.bgcolor = "#cacaca";//"white";
                        break;
                    case 49://default
                        tty.font.bgcolor = "#cacaca";//"black";
                        break;
                    case 90 :
                        tty.font.color = "black";
                        break;
                    case 91:
                        tty.font.color = "red";
                        break;
                    case 92:
                        tty.font.color = "green";
                        break;
                    case 93:
                        tty.font.color = "yellow";
                        break;
                    case 94:
                        tty.font.color = "blue";
                        break;
                    case 95:
                        tty.font.color = "magenta";
                        break;
                    case 96:
                        tty.font.color = "cyan";
                        break;
                    case 97:
                        tty.font.color = "white";
                        break;
                    case 100:
                        tty.font.bgcolor = "black";
                        break;
                    case 101:
                        tty.font.bgcolor = "red";
                        break;
                    case 102:
                        tty.font.bgcolor = "green";
                        break;
                    case 103:
                        tty.font.bgcolor = "yellow";
                        break;
                    case 104:
                        tty.font.bgcolor = "blue";
                        break;
                    case 105:
                        tty.font.bgcolor = "magenta";
                        break;
                    case 106:
                        tty.font.bgcolor = "cyan";
                        break;
                    case 107:
                        tty.font.bgcolor = "white";
                        break;
                    default:
                        tty.fn.defaultFont(tty);
                    }
                    N = r.exec(re_a[1]);
                }
            }, 
	    'rebuildRows' : function (tty, first) {
		first = typeof first == 'undefined' ? 0 : first;
		tty.cache.rows = "rows" in tty.cache ? tty.cache.rows : new Array();
		tty.cache.rows.length = first;
		$(".ttyrow", tty.node.ttyel).each(
                    function(i,el){
			if (i>=first) {
			    tty.cache.rows.push(el);
			    tty.fn.rebuildCols(tty, tty.cache.rows.length);
			}
                    }
                );
            },
            'eFindRow' : function (tty, sy) {
                while(tty.cache.rows.length<sy){
                    var row = $(tty.emptyRow).appendTo(tty.node.ttyel);
                    tty.cache.rows.push(row[0]);
                    tty.fn.rebuildCols(tty, tty.cache.rows.length);
                    if(tty.cache.rows.length - tty.max_y > 0){
                        $('.ttycell',
                          $(tty.cache.rows[tty.cache.rows.length - tty.max_y - 1])
                          .html(
                              '<div class="ttystr">'
                                  +$(tty.cache.rows[tty.cache.rows.length - tty.max_y - 1])
                                  .text().replace(RegExp('(.{' + tty.max_x + '})(?=.)','gm'),"$1<br/>")+"</div>"));
//.css("width",$(tty.cache.rows[tty.cache.rows.length - tty.max_y - 1]).text().length+'ex');
                        tty.cache.rows[tty.cache.rows.length - tty.max_y - 1].nsgCols = null;
                        tty.node.ttyel.scrollTop=tty.node.ttyel.scrollHeight - tty.node.ttyel.clientHeight;
                    }
                }
                return tty.cache.rows[sy-1];
            },
            'rebuildCols' : function (tty, sy) {
                var row =  tty.fn.eFindRow(tty, sy);
                row.nsgCols = "nsgCols" in row ? row.nsgCols : new Array();
		row.nsgCols.length = 0;
                 $(".ttycell, .ttystr",row).each(
                    function(i,el){
                       row.nsgCols.push(el);
                    }
                );
            },
	    'eFindCell' : function (tty, sy, sx) {
		if (sx > tty.max_x)
		    return null;
		var row =  tty.fn.eFindRow(tty, sy);
                if (row.nsgCols != null ) {
		    if(row.nsgCols.length<sx){
			while(row.nsgCols.length<sx){
			    row.nsgCols.push($("<div class='ttycell'>&nbsp;</div>").appendTo(row)[0]);
			}
			tty.fn.rebuildCols(tty, sy);
		    }
		    return row.nsgCols[sx - 1];
                } else return null;
            },
            'eSet' : function (tty, c) {
		tty.fn.wrap(tty);
                var cell = tty.fn.eFindCell(tty, tty.y + tty.offset, tty.x);
                $(cell).html(htmlEncode(c).replace(/ /g,'&nbsp;'));
		tty.x = tty.x + 1;
		
//	tty.node.ttyel.scrollTop=tty.node.ttyel.scrollHeight;
                if("white" !=  tty.font.inverse ? tty.font.bgcolor : tty.font.color) {
                    $(cell).css('color', tty.font.inverse ? tty.font.bgcolor : tty.font.color);
                }
                if("black" !=  tty.font.inverse ?  tty.font.color :tty.font.bgcolor) {
                    $(cell).css('background-color', tty.font.inverse ?  tty.font.color :tty.font.bgcolor);
                }
                if(tty.font.underlined)
                     $(cell).css('text-decoration',  tty.font.underlined ? 'underline' : 'none');
                if(tty.font.bold)
                     $(cell).css('font-weight',tty.font.bold ? 'bold' :'normal');
                //debug  tty.node.setCursor();
//		tty.node.ttyel.scrollTop=tty.node.ttyel.scrollHeight;

            }
        }
    };
    this.fastRestart = false;
    this.ctrlKey = false
    this.receiving_elementID = receiving_elementID;
    /***********************************************************************
       Creates XMLHTTPRequest object
    ***********************************************************************/
    this.create_transfer_object = function()
    {
        var Request = false;

        if (window.XMLHttpRequest)
        {
	          /*Gecko- Safari, Konqueror*/
	          Request = new XMLHttpRequest();
        }
        else if (window.ActiveXObject)
        {
	          /*Internet explorer*/
	          try
            {
                Request = new ActiveXObject("Microsoft.XMLHTTP");
            }
	          catch (CatchException)
            {
                Request = new ActiveXObject("Msxml2.XMLHTTP");
            }
        };

        if (!Request)
        {
	          alert("Can`t create XMLHttpRequest");
        };

        return Request;
    }
  
    /*******************************************************
     REGEXP FUNCTIONS
    *******************************************************/
    this.escRegExp = [{
        'r':/^(\x1b[\x00-\x1f]*)/,
        'n':[
            {
                'r':/^([\[])/,
                'n':[
                    { 'r':/^([!]p)/,/*  TODO reset terminal*/
		      'f': this.tty.fn.resetTerminal
		    },
                    { 'r':/^([0-9;]*)?m/,
                      //'r':/^((;)?([0-9][0-9]?)?)*m/,
                      // set color params
                      'f': this.tty.fn.setColorParams
                    },
                    { 'r':/^(0?c)/,//Requests a Report Device Code response from the device.
                      // 'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b[00c");}},
                      'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b[?1;0c");}},

                    { 'r':/^(5n)/,//Requests a Report Device Status response from the device.
                      'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b[0n");}},
                    { 'r':/^(6n)/,//Requests a Report Cursor Position response from the device.
                      'f':function (tty, re_a) { tty.node.buffer_next_data("\x1b["+(tty.y)+";"+(tty.x)+"R");}},
                    
		    { 'r':/^([0-9]+)?(;([0-9]+)?)?[Hf]/,
                      // Set cursor position
                      'f': this.tty.fn.setCursorPos
                    },
                    { 'r':/^([0-9]+)?A/,
                      // Move cursor UP
                      'f': this.tty.fn.moveCursorUp
                    },
                    { 'r':/^([0-9]+)?B/,
                      // Move cursor DOWN
                      'f': this.tty.fn.moveCursorDown
                    },
                    { 'r':/^([0-9]+)?C/,
                      // Move cursor right
                      'f': this.tty.fn.moveCursorRight
                    },
                    { 'r':/^([0-9]+)?d/,
                      // Move cursor to row
                      'f': this.tty.fn.moveCursorToRow
                    },
                    { 'r':/^([0-9]+)?D/,
                      // Move cursor left
                      'f': this.tty.fn.moveCursorLeft
                    },
                    { 'r':/^([0-9]+)?[G`]/,
                      // Move cursor to column
                      'f': this.tty.fn.moveCursorToColumn
                    },
                    { 'r':/^([0-9]+)?L/,
                      // Insert Line
                      'f': this.tty.fn.insertLine
                    },
                    { 'r':/^([0-9]+)?M/,
                      // Delete Line
                      'f':this.tty.fn.deleteLine
                    },
                    { 'r':/^([0-9]+)?P/,
                      // Delete Char
                      'f':this.tty.fn.deleteChar
                    },
                    { 'r':/^s/, //save cursor pos
                      'f':this.tty.fn.saveCursorPos
                    },
                    { 'r':/^u/, //restore cursor pos
                      'f':this.tty.fn.restoreCursorPos
                    },
                    { 'r':/^(([1-9][0-9]?)?(;([1-9][0-9]?[0-9]?))?r)/,
                      // Set scroll region
                      'f': this.tty.fn.setScrollRegion
                    },
                    { 'r':/^([0-9]+)?g/,
                      'f':this.tty.fn.clearTab
		    },
                    { 'r':/^[0]?K/,
                      'f': this.tty.fn.clearLineFromPos
                    },
                    { 'r':/^1K/,
                      'f': this.tty.fn.clearLineBeforePos
                    },
                    { 'r':/^2K/,
                      'f': this.tty.fn.clearLine
                    },
                    { 'r':/^[0]?J/,
                      'f':  this.tty.fn.clearScreenFromPos
                    },
                    { 'r':/^1J/,
                      'f':  this.tty.fn.clearScreenBeforePos
                    },
                    { 'r':/^2J/,
                      'f': this.tty.fn.clearScreen
                    },
                    { 'r':/^([1-9][0-9]?)?X/,
                      'f': this.tty.fn.eraseChar
                    },
                    {
                        'r':/^[?][0-9;]*c/,
                        'f':this.tty.fn.nil
                    },
		    
		    {'r':/^20([lh])/,
		     'f':function(tty, re_a) { 
			  tty.lnm = re_a[1]=='h';
		     }
		    },
		    {'r':/^[?]([0-9]*)([lh])/,
		     'f':function (tty, re_a) {
			 var n = Number(re_a[1]);
			 var m = re_a[2];
			 var l = 'l';
			 var h = 'h';
			 switch(n) {
			 case 0:
			     break;
			 case 1://Set cursor key to application/cursor h/l
			     break;
			 case 2://Set VT52 (versus ANSI)  l
			     break;
			 case 3://Set number of columns to 132/80 h/l
			     tty.max_x = m==h ? 132 : 80;
			     tty.fn.resetTerminal(tty);
			     break;
			 case 4://Set jump smooth/scrolling  h/l
			     break;
			 case 5://Set reverse/normal video on screen h/l
			     tty.inverse = m==h;
			     break;
			 case 6://Set origin to relative/absolute h/l
			     tty.relative = m==h;
			     break;
			 case 7://Line WRAP ON/OFF h/l
			     tty.wrap = m==h;
			     break;
			 case 8://Set/Reset auto-repeat mode h/l
			     break;
			 case 9://Set/Reset interlacing mode h/l
			     break;
			     
			 }
		     }
		    },
		    
		    { 'r':/^(O|N)/,
		      // Application/Numeric  keypad mode
		      'f': this.tty.fn.nil
		    },	  
                ]
            },
	    

            { 'r':/^([#]8)/,
              'f':this.tty.fn.fillScreenWithE
            },
            { 'r':/^([=><])/,
              'f':this.tty.fn.nil
            },
            { 'r':/^(c)/,/*  TODO reset terminal*/
              'f': this.tty.fn.resetTerminal
            },
            { 'r':/^([\(])[\x40-\x7e]/, /*  TODO set default font*/
              'f':this.tty.fn.nil
            },
            { 'r':/^([\)])[\x40-\x7e]/, /*  TODO set alternate font*/
              'f':this.tty.fn.nil
            },
            { 'r':/^([\)\(])[BO0UK]/, /*  TODO Start sequence defining G0/G1 character set*/
              'f':this.tty.fn.nil
            },
            { 'r':/^7/, //save cursor pos
              'f': this.tty.fn.saveCursorPos
            },
            { 'r':/^8/, //restore cursor pos
              'f': this.tty.fn.restoreCursorPos
            },
            { 'r':/^D/, // Down one line, scrolling if neccessary
              'f':  function (tty) {
		  tty.fn.lineFeed(tty);
	      }
            },
            { 'r':/^E/, // return + linefeed    Next line.
              'f': function (tty) { 
		  tty.fn.Return(tty);
		  tty.fn.lineFeed(tty);
	      }
            },
            { 'r':/^M/, // Up one line, scrolling if neccessary
              'f': function (tty) { 
		  
		  if(tty.y > tty.offset + tty.scroll.top) {
		      tty.y = tty.y-1; 
		  } else {
		      tty.fn.scrollRegionDown(tty, 1);
		  }
	      }
            },
            { 'r':/^H/, // Set TAB position
              'f':function (tty) { 
		  tty.tab[tty.x] = true;
	      }
            },
            { 'r':/^Z/, // ident
//              'f':function (tty) {  tty.node.buffer_next_data("\x1b\\Z");}
              'f':function (tty) {  tty.node.buffer_next_data("\x1b[?1c");}
            }/*,
            {
                'r':/^(.*)/, // Unknown escape
                'f':function (tty, re_a) { alert(re_a[1]); }
            }*/
        ]
    }];

    this.runRegexp = function(data, escRegExp_a) {
        for (var i=0; i<escRegExp_a.length; i++ ) {
            if (data.search(escRegExp_a[i].r) == 0) {
                var regRes = escRegExp_a[i].r.exec(data);
                if (escRegExp_a[i].f) {
                    escRegExp_a[i].f(this.tty, regRes);
                    return regRes[0].length;
                } else if (escRegExp_a[i].n) {
                    var fRes = this.runRegexp(data.substring(regRes[0].length), escRegExp_a[i].n);
                    return  ( fRes >0 ? regRes[0].length + fRes : 0);
                };
            }
        }
        return 0;
    }

   
    /***********************************************************************
       callback function whenever html data is ready
    ***********************************************************************/
    this.get_html_state = function()
    {
        if (this.xhr_out.readyState == 4)
        {
            //alert("status:"+this.xhr_out.status);
            if ((this.xhr_out.status!=200)
                &&(this.xhr_out.status!=204)
                &&(this.xhr_out.status!=0)
                &&(typeof(this.xhr_out.status) != 'undefined'))
                this.error();
            if (this.xhr_out.status==200 || this.xhr_out.status==204 || (typeof(this.xhr_out.status) == 'undefined')) {
                this.initCounter = this.initCounter + 1;
                //alert('responseText:'+zeroOrText(this.xhr_out.responseText));
                hiddenDiv.innerHTML = zeroOrText(xmlTransform(this.xhr_out));
                //alert('after\10'+hiddenDiv.innerHTML)
                execScripts(hiddenDiv);
                var response = htmlDecode(getText(hiddenDiv));
                if(this.initCounter==1) {
                    $(this.element).html("<img title=\"header=[webJobs] body=[1]\" class=\"ttyactivity ttyactivityon\" src=\"/loading.gif\" alt=\"\"><div class='ttyscroll'><div class='ttybody' tabindex=0 ><div class='ttyrow0'></div><br/></div></div>");
		    this.ttyel = $("div.ttybody" , this.element)[0];
                    $('head style.tty').remove();
                    $('head').append("<style  type='text/css' class='tty'>\ndiv.ttycell {\nwidth:"+($("#ttysizer").width())+"px;\nline-height:"+$("#ttysizer").height()+"px;\n}\ndiv.ttybody {\n height:"+($("#ttysizer").height()*24.5)+"px;}\n</style>");

                    for (var i=0; i<this.tty.max_x; i++ ) {
                        $("div.ttyrow0", this.ttyel).append("<div class='ttycell'>&nbsp;</div>");
                    }
           
                    this.session_id  = ":)";
                    this.activity_img = $("img.ttyactivity", this.element)[0];
                    this.tty.fn.resetTerminal(this.tty);

                    this.handle_data(response);
                    var obj = this;
                    this.ttyel.onkeypress = function(e) {
			if (obj.running) return obj.key_event(e); 
		    };
                    this.ttyel.onkeydown  = function(e) { 
			if (obj.running) return obj.special_key_event(e); 
		    };
		    this.ttyel.onscroll = function() { 
			if (obj.tty) {
			    obj.tty.autoscroll = obj.ttyel.scrollTop == (obj.ttyel.scrollHeight - obj.ttyel.clientHeight);
			}
		    };
		    
                    $(this.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
		    this.timeout(
			function() {
                            if(webTtyTaskArray[obj.receiving_elementID] != null)
				webTtyTaskArray[obj.receiving_elementID].get_data();
			},
			10);
		    
		    this.ttyel.focus();
                }
                else {
                    handle_data(response);
                }
            }
        }
    }

    /***********************************************************************
       Request a new process and HTML data from server
    ***********************************************************************/
    this.get_html = function()
    {
        var obj = this;
        var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
        //      if (null == this.xhr_out)
	this.tty.state = TTY_STATE_RUNNING;
        this.xhr_out = SendRequestAsync('POST',
                                        PAGENAME,
                                        'command=tty&path={'+path+'}&value=&',
                                        function() { if (obj.running) obj.get_html_state(); },
                                        "retry"
                              );
    }




    /***********************************************************************
       callback function whenever process data is ready
    ***********************************************************************/
    this.get_data_state = function()
    {
        if (this.xhr_out.readyState == 4)
        {
	    $(this.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
            if ((this.xhr_out.status!=200)&&(this.xhr_out.status!=204)&&(this.xhr_out.status!=0)
                &&(typeof(this.xhr_out.status) != 'undefined'))
                this.error();
            hiddenDiv.innerHTML = zeroOrText(xmlTransform(this.xhr_out));
            execScripts(hiddenDiv);
            var data = htmlDecode(getText(hiddenDiv));
            if ( ! this.handle_data(data))
            {   

		this.handle_data("", true);
                this.session_id=null;
                if (undefined != this.webtty_exit)
                    this.webtty_exit(0);
                this.cleanup();
            }
            else
            {
                var obj =this;
                this.timeout(
		    function() {
               		$(obj.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
			if(webTtyTaskArray[obj.receiving_elementID] != null)
                            webTtyTaskArray[obj.receiving_elementID].get_data();
                    },
		    1000);
                
            }
        }
    }


    /***********************************************************************
       Request new data from the server
    ***********************************************************************/
    this.get_data = function()
    {
        if (this.timerID != null) {
            clearTimeout( this.timerID);
            $(this.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
        }
        this.timerID = null;
	var obj = this;
	if (this.tty.state < TTY_STATE_CLOSING && this.kbd_buffer =='' && this.tty.data.length>0){
	    this.timeout(
		function() {
               	    $(obj.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
		    if(webTtyTaskArray[obj.receiving_elementID] != null)
                        webTtyTaskArray[obj.receiving_elementID].get_data();
                },
		300);
	    return;
	}
	
        var out;
        var k;
        var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
        k = this.kbd_buffer.search(/[%][%][%]/);
        if(k >= 0) {
	    this.tty.state = TTY_STATE_CLOSING;
	    $("span.ttynote", document.getElementById(receiving_elementID).parentNode).html(getHelp("webTtyClosing")).removeClass('blue yellow red green').addClass("yellow");
            out = "";
            this.kbd_buffer = '';
	    this.tty.data = '';
	} else {
            out = "&value="+this.kbd_buffer;
	};
	if (this.tty && this.tty.state < TTY_STATE_CLOSED ) {
	    $(this.activity_img).removeClass("ttyactivityoff").addClass("ttyactivityon");
            this.xhr_out = SendRequestAsync('POST',
					    PAGENAME,
					    'command=tty&path={'+path+'}&'+ out + '&',
					    function() { return obj.get_data_state(); },
                                            "retry"
					   );
	}
        this.kbd_buffer = '';
    }


    /***********************************************************************
       exit page. Remove handles and terminate session
    ***********************************************************************/
    this.cleanup = function()
    {
        this.running     = false;
        if (this.timerID) {
            clearTimeout( this.timerID);
            $(this.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
        }
        this.timerID = null;
        webTtyTaskArray[this.receiving_elementID] = null;
        if (this.xhr_kbd)
        {
            this.xhr_kbd.onreadystatechange = function() {};
            this.xhr_kbd.abort();
            delete this.xhr_kbd;
        }
        if (this.xhr_out)
        {
            this.xhr_out.onreadystatechange = function() {};
            this.xhr_out.abort();
            delete xhr_out;
        }
        if (null != this.session_id)
        {
            var path = encodeURIComponent(buildPath(findParentByTag(this.element, "LI", 10), ","));
            var exit_req = SendRequestAsync('POST',
                                            PAGENAME,
                                            'command=tty&path={'+path+'}&',
                                            function() {}
                                           );
            delete exit_req;
            this.session_id=null;
        }
        if (null != this.tty)
            this.tty = null;
    }
    
    /***********************************************************************
       handles process data
    ***********************************************************************/
    this.handle_data = function(data, force)
    {	var tty = this.tty;
	if (! tty) 
	    return false;
	var max_i = 100;
	var i = force ? max_i + 10 : 0; 
	var GOOD = true;
	if (data == '_C_L_O_S_E_D_') {
	    data = '';
	    if (this.tty.state < TTY_STATE_CLOSED ) {
		$("span.ttynote", document.getElementById(receiving_elementID).parentNode).html(getHelp("webTtyClosed")).removeClass('blue yellow green red').addClass("red");
	    }
	    GOOD = false;
	    this.tty.state = TTY_STATE_CLOSED;
	    return false;
	}
	if(this.tty.state >= TTY_STATE_CLOSING) {
	    tty.data = '';
	} else {
	    tty.data = tty.data + data;
	}
        while (tty.data.length > 0 && i<max_i) {
	    i ++;
            var c = tty.data.charAt(0);
            // try {
	    if(c=='\x1b') {
		var CSIre = /^\x1b[\x00-\x1f]*[[][\x00-\x1f0-9;?>!$'*&]*[@A-MPSTXZ`a-il-np-z{}|~]/;
		var clearCSIre = /^[\x00-\x1f]*\x1b[^\x00-\x1f]*$/
		    var CSI = tty.data.match(CSIre);
		CSI = CSI ? CSI[0] : null;
		var corCSI = CSI;
		
		while (corCSI && ! clearCSIre.test(corCSI)) {
		    corCSI = corCSI.replace(/\x1b([^\x00-\x1f]*)([\x00-\x1f]*)(.*)/, '$2\x1b$1$3');
		}
		if (corCSI != CSI) 
		    tty.data = corCSI + tty.data.substring(corCSI.length);
	    }
	    c = tty.data.charAt(0);
            var esc = (c=='\x1b' ?  this.runRegexp(tty.data, this.escRegExp) : 0);
            if (esc>0) {
                tty.data = tty.data.substring(esc);
            } else {
                if(c=='\t') {
                    c=" ";
                    tty.data = ' '+tty.data.substring(1);
                    var i = tty.fn.getTabShift(tty);
                    while (i>0) {
                        i--;
                        tty.data = ' '+tty.data;
                    }
                }
		/*            if(c==' ')
                              c = '&nbsp;';
		*/
                if(c=='\n' /* 0x0A  */ || c=='\x0B' || c=='\x0C') {
                    tty.data = tty.data.substring(1);
		    //	tty.fn.Return(tty);
		    tty.fn.lineFeed(tty);                                       
                }else if(c=='\x08') {
                    tty.data = tty.data.substring(1);
                    tty.fn.moveCursorLeft(tty, [1,1]);		    
                }else if(c=='\r' /* 0x0D */) {
                    tty.data = tty.data.substring(1);
		    tty.fn.Return(tty);
		    //	tty.fn.lineFeed(tty);                       
                }else if(c>='\x00' && c<='\x1F' ) {
                    tty.data = tty.data.substring(1);
                } else {
                    tty.data = tty.data.substring(1);
                    tty.fn.eSet(tty, c);                      
                }
            };

            if(1) {
                //!!!!!! DEBUG ONLY
                //this.eelement.scrollTop=this.eelement.scrollHeight;
            }
            /*  } catch(err) {
                tty.data = tty.data.substring(1);
                tty.x = tty.x+1;
                if(tty.x > tty.max_x && ! tty.wrap) {
                tty.x = tty.max_x;
                }
		}*/

        };  
	if (this.tty.autoscroll)
	    this.ttyel.scrollTop = this.ttyel.scrollHeight - this.ttyel.clientHeight;
        this.setCursor();
	
	if(i>=max_i) {
	    setTimeout(function() {tty.node.handle_data('');}, 1);
	}
	return GOOD;
    }

    this.ms = function(){
	return new Date().getTime();
    }

    this.timeout = function(f,p){
	p = this.fastRestart>0 ? 100 : p;
	this.fastRestart --;
	this.fastRestart = this.fastRestart <0 ? 0 : this.fastRestart;
	if(this.timerID){
	    clearTimeout(this.timerID);
	    this.timerID = null;
	}
	this.timerPeriod = p;
	this.timerStart = this.ms();
        this.timerID = setTimeout(f, p);
    }
   
    /***********************************************************************
      buffer data and send to server
    ***********************************************************************/
    this.buffer_next_data = function(cmd)
    {
        this.kbd_buffer += cmd;
        if (this.timerID != null && ((this.timerStart+this.timerPeriod)-this.ms() > 100)) {
            $(this.activity_img).removeClass("ttyactivityon").addClass("ttyactivityoff");
	    var obj = this;
            this.timeout(function() {
		webTtyTaskArray[obj.receiving_elementID].get_data();
            }, 10);
	}
	this.fastRestart ++;
	this.fastRestart = this.fastRestart >3 ? 3 : this.fastRestart;
    }



  /***********************************************************************
       handles normal keypresses
  ***********************************************************************/
  this.key_event = function(e)
    {
        var key;
        var ch;
	var ewe = false;
        if (!e) {
	    e = window.event;
	    ewe = true;
	}
        if (e.char)
            key = e.char.charCodeAt(0);
        else if (e.keyCode)
            key = e.keyCode;
        else if (e.which)
            key = e.which;
        if ( !ewe && typeof(e.charCode)=="undefined")
            return true;
        /*handle escape symbols for Ctrl X*/
        if ((String.fromCharCode(key) == "^") && !this.ctrlKey) {
            this.ctrlKey = true;
            return false;
        }
        if (this.ctrlKey) {
            this.ctrlKey = false;
            if ( (key > 64) && (key < 96) && (key != 94) )
                key = key - 64;
        }
        
        if (key==13 || key == 10 || key == 9 || key == 8) {
            return false;
        }

        this.buffer_next_data(fullURIEncode(key));
        return false;
    }


  /***********************************************************************
       handles special keypresses
  ***********************************************************************/
  this.special_key_event = function (e)
    {
      if (!e) 
	  e = window.event;
      if (e.keyCode) key = e.keyCode;
      else if (e.which) key = e.which;

     
      if (key==9) //tab
        { 
          this.buffer_next_data(escape("\t")); 
          return false; 
        }
      else if (key==8) // back space
        { 
          this.buffer_next_data(escape("\010")); 
          return false; 
        }
      else if (key == 40) 
        { 
          this.buffer_next_data(escape("\033[B")); 
          return false; 
        }
      else if (key == 38)
        { 
          this.buffer_next_data(escape("\033[A")); 
          return false; 
        }
      else if (key == 37) //left
        { 
          this.buffer_next_data(escape("\033[D")); 
          return false; 
        }
      else if (key == 39) //right
        { 
          this.buffer_next_data(escape("\033[C")); 
          return false; 
        }
	else if (key == 46) //del
        { 
          this.buffer_next_data(escape("\033[X")); 
          return false; 
        }
	else if (key == 33) // PageUp
        { 
          this.buffer_next_data(escape("\033[V")); 
          return false; 
        }
	else if (key == 34) // PageDown
        { 
          this.buffer_next_data(escape("\033[U")); 
          return false; 
        }
	else if (key == 36) // Home
        { 
          this.buffer_next_data(escape("\033[H")); 
          return false; 
        }
	else if (key == 35) // End
        { 
          this.buffer_next_data(escape("\033[F")); 
          return false; 
        }
      else if (key == 13 || key == 10)
        {
          this.buffer_next_data(escape("\r"));
          this.cr = true;
          return false; 
        }
    }


  /***********************************************************************
       handles errors
  ***********************************************************************/
    this.error = function()
    {
	if (undefined != this.webtty_exit)
            this.webtty_exit(1);
	if (this.element)
            this.element.innerHTML="Error";
	this.cleanup();
    }


    this.element =  document.getElementById(receiving_elementID);
    if (null == this.element)
	return this.error();

    this.running     = true;
    this.session_id  = null;
    this.output_area = null;
    this.kbd_buffer  = '';
    this.webtty_exit = exit_func;
    this.timerID = null;
    this.timerStart = 0;
    this.timerPeriod = 1000;
    this.initCounter = 0;
    this.fastRestart = false;
    webTtyTaskArray[receiving_elementID] = this;
}


function init_webtty_object(objID, exit_handler) {
  /*if (undefined != exitFunc) 
    exit_handler = exitFunc;
    else
    exit_handler =  function webtty_exit()
    {
    var div_element = obj;
    div_element.innerHTML = 'Refresh page to get a new virtual machine';
    };*/
  a = new webtty(objID, exit_handler);
  a.get_html();
  return a;
}
